import utils
import types
import sys
import random as rand
import copy
from numpy import *

def print_mat(mat):
    for i in range(len(mat)):
        for j in range(len(mat[0])):
            print "%.04f" % mat[i][j],
        print ""

class Motif:
    
    def __init__(self, fileIn):
        
        self.name = None
        self.ID = None
        self.FSM = None
        self.n = None
        self.numSamples = None
        self.matFSM = None
        self.matPSSM = None
        self.seq = []

        self.empty = False
        
        if(fileIn != None):
            self.readFromFile(fileIn)
        else:
            print "Wrong usage for Motif.__init__()"

        self.intuitM = zeros((16,(self.n * (self.n -1))/2)) # Matrix for the membership degree "\mu"
        self.intuitV = zeros((16,(self.n * (self.n -1))/2)) # Matrix for the non-membership degree "\nu"
    
        self.computeIntuitionisticMotif()

    def __len__(self):
        return self.n
    
    def calculatePSSMFromFSM(self):
        self.matPSSM = array(self.matFSM,float)
        self.matPSSM /= self.numSamples

    def calculateFSMFromPSSM(self,numSamples=None):
        if numSamples == None:
            numSamples = self.numSamples
        
        self.numSamples = numSamples

        if self.numSamples is None:
            self.numSamples = 50

        self.matFSM = array(self.matPSSM*self.numSamples).astype(int)
        for i in range(self.n):
            s = sum(self.matFSM[i,:])
            ind = self.matFSM[i,:].tolist().index(max(self.matFSM[i,:]))
            self.matFSM[i,ind] += self.numSamples - s

    def pseudoCountFSM(self):
        if not all(self.matFSM):
            self.matFSM += 1
            self.numSamples += 4
            self.calculatePSSMFromFSM()

    def pseudoCountPSSM(self):
        if not all(self.matPSSM):
            
            self.matPSSM += 0.000001
            self.calculateFSMFromPSSM()

                         

    def ICmatrix(self):
    
        self.pseudoCountPSSM()
        matAux = self.matPSSM * (log(self.matPSSM)/log(2))      
        return matAux
    
    def print_to_file(self, out_file, mode='FSM', force = False):
        if not self.empty or force:
            fo = open(out_file, 'w')
            if mode == 'FSM':
                self._print_to_file_fsm(fo)
            elif mode == 'PWM':
                self._print_to_file_pwm(fo)
            else:
                print "Warning: Unknown file mode. Using FSM mode instead."
                self._print_to_file_fsm(fo)

            fo.close()
        else:
            print "Error: Empty motif. Not writing it."

    def _print_to_file_fsm(self, out_handle):
        self._write_header_to_file(out_handle)
        self._write_matrix_to_file(self.matFSM, out_handle)
        self._write_sequences_to_file(self.seq, out_handle)
        
    def _write_header_to_file(self, out_handle):
        header = '>%s\n' % self.ID
        out_handle.write(header)

    def _write_matrix_to_file(self, matrix, out_handle):
        header = 'A\tC\tG\tT\n'
        out_handle.write(header)

        for row in matrix:
            out_handle.write('\t'.join([str(i) for i in row]) + '\n')

    def _write_sequences_to_file(self, sequences, out_handle):
        header = 'SEQUENCES\n'
        out_handle.write(header)

        for s in sequences:
            out_handle.write(s+'\n')


    def _print_to_file_pwm(self, out_handle):
        self._write_header_to_file(out_handle)
        self._write_matrix_to_file(self.matPSSM, out_handle)
        self._write_sequences_to_file(self.seq, out_handle)
                
    def readFromFile(self,fileIn):

        if type(fileIn) == types.StringType: fileIn = open(fileIn)
        
        header = fileIn.readline().split()
        AC = header[0].replace('>','')
        name = AC
        acgt_line = fileIn.readline()
        if acgt_line != 'A\tC\tG\tT\n':
            print "Warning: Error likely to happen, no ACGT header in file."
            
        mat = []
        col = fileIn.readline().split()
        pos = fileIn.tell()
        while(col and col[0] != 'SEQUENCES'):
            mat.append([float(s) for s in col])
            pos = fileIn.tell()
            col = fileIn.readline().split()
                
        col = fileIn.readline().upper().replace('\n','')
        
        
        while col and not utils.EOF(fileIn) and col[0] != '>':
            self.seq.append(col)

            col = fileIn.readline().upper().replace('\n','')
            
        
        mat = array(mat,float)
        
        if self.FSM == None:
            self.FSM = utils.decideFSMorNot(mat)
        
        if self.FSM == True:
            self.numSamples = self.computeNumSamples(mat)
            self.matFSM = mat.astype(int)
            self.n = mat.shape[0]
            self.calculatePSSMFromFSM()
            self.name = name
            self.ID = AC

        elif self.FSM == False:
            self.matPSSM = array(mat,float)
            self.n = mat.shape[0]
            self.calculateFSMFromPSSM()
            self.name = name
            self.ID = AC

        if self.matFSM != []:
            # added by CNL: Generating sequences
            if self.seq == []:
                self.generate_sequences()
        else:
            print "Warning, empty motif: %s" % self.ID
            self.empty = True

        
    def generate_sequences(self):
        self.seq = []

        base_index = {'A':0, 'C':1, 'G':2, 'T':3}

        if self.validate_fsm:
            num_sequences = sum(self.matFSM[0])

            counts_matrix = copy(self.matFSM)

            for i in range(num_sequences):
                new_sequence = self.generate_one_sequence(counts_matrix)

                self.seq.append(new_sequence)
                
        else:
            print "Error generating sequences: Not all rows of fsm matrix sum the same value."


    def generate_one_sequence(self, counts_matrix):
        bases = ['A','C','G','T']
        result = ''

        random.seed()

        for row in counts_matrix:
            indexes = [i for i in range(len(row)) if row[i] > 0]

            chosen_index = rand.choice(indexes)

            row[chosen_index] -= 1
            result += bases[chosen_index]

        return result


    def validate_fsm(self):
        fsm_ok = True

        row_sum = sum(self.matFSM[0])
        for i in range(1,len(self.matFSM)):
            new_sum = sum(self.matFSM[i])
            if row_sum != new_sum:
                fsm_ok = False

        return fsm_ok

    def computeNumSamples(self,mat):
        numSamples = 0
        for i in range(len(mat)):
            if numSamples < sum(mat[i,:]):
                numSamples = sum(mat[i,:])
        return numSamples

    
    def computeIntuitionisticMotif(self):
        t = float(len(self.seq))
        F = zeros((16,(self.n * (self.n -1))/2))
        a = 0.000001 * 0.000001

        dposmatFSM = {}
        dposmatFSM['A'] = 0
        dposmatFSM['C'] = 1
        dposmatFSM['G'] = 2
        dposmatFSM['T'] = 3

        drow = {}
        drevrow = {}
    
        conti = 0
        for i in ['A', 'C', 'G', 'T']:
            for j in ['A', 'C', 'G', 'T']:
                drow[str(i)+str(j)] = conti
                drevrow[conti] = str(i)+str(j)
                conti += 1

        dcol = {}
        drevcol = {}
        conti = 0
        for i in range(self.n):
            for j in range(i+1,self.n):
                dcol[(i,j)] = conti
                drevcol[conti] = (i,j)
                conti += 1
        
        for s in self.seq:
            for indexi, posi in enumerate(s[0:-1]):
                for indexj, posj in enumerate(s[indexi+1:]):
                    F[drow[str(posi)+str(posj)], dcol[(indexi,indexj+indexi+1)]] += 1
    
        self.intuitM = (F /t) + a

        for indexi in range(self.n):
            for indexj in range(indexi+1,self.n):
                for posi in ['A','C','G','T']:
                    for posj in ['A','C','G','T']:
                        izq = self.matPSSM[indexi][dposmatFSM[str(posi)]]
                        dcha = self.matPSSM[indexj][dposmatFSM[str(posj)]]
                        tot1 = (izq + dcha) / 2.
                        oldM = self.intuitM[drow[str(posi)+str(posj)],dcol[(indexi,indexj)]]
                        self.intuitM[drow[str(posi)+str(posj)],dcol[(indexi,indexj)]] += (1-oldM)*tot1                      
                
        ic = self.ICmatrix()

        for indexi in range(self.n):
            for indexj in range(indexi+1,self.n):
                for posi in ['A','C','G','T']:
                    for posj in ['A','C','G','T']:
                        izq = (2 + ic[indexi,dposmatFSM[str(posi)]]) / 2
                        dcha = (2 + ic[indexj, dposmatFSM[str(posj)]]) / 2
                        tot1 = (izq + dcha) / 2
                        tot2 = tot1 * (1- self.intuitM[drow[str(posi)+str(posj)],dcol[(indexi,indexj)]])
                        self.intuitV[drow[str(posi)+str(posj)],dcol[(indexi,indexj)]] = tot2


    # for compatibility in scan
    def score(self, seq):
        return self.SC_intuit(seq)

    def SC_intuit(self,seq):
        drow = {}
        drevrow = {}
        
        conti = 0
        for i in ['A','C','G','T']:
            for j in ['A','C','G','T']:
                drow[str(i)+str(j)] = conti
                drevrow[conti] = str(i)+str(j)
                conti += 1
    
        dcol = {}
        drevcol = {}
        conti = 0
        for i in range(self.n):
            for j in range(i+1,self.n):
                dcol[(i,j)] = conti
                drevcol[conti] = (i,j)
                conti += 1
    
        
        A = [self.intuitM.transpose(),self.intuitV.transpose()]
        
        
        simsNorm = []
        for indexi,posi in enumerate(seq[0:-1]):
            for indexj,posj in enumerate(seq[indexi+1:]):
                maxSim,minSim = utils.computeMaxMin(A,dcol[(indexi,indexj+ indexi +1)])
                mu = A[0][dcol[(indexi,indexj+ indexi +1)]][drow[posi+posj]]
                maxV =  max(A[1][dcol[(indexi,indexj+ indexi +1)]])
                v = maxV - A[1][dcol[(indexi,indexj+ indexi +1)]][drow[posi+posj]]
                curSim = mu * v
                curSimNorm = (curSim - minSim) / (maxSim -minSim)
                simsNorm.append(curSimNorm)

        return average(simsNorm)
